#include <winuser.h>
#include <windows.h>
#include <math.h>
#include <string.h>
#include <stdlib.h>
#include "Dotzres.h"

#define dot_historyy 1024//limiting value for history. Actual amount of memory used depends on it rather than value from .ini file
#define show_fps 0       //1:show fps 0:don't show fps

char lpIniFileName[] = "Dotz.ini";
char lpIniFileNameAndPath[256];       // storage area for ini file name + path
char lpIniSection[64];                // storage area for section name
char lpIniKeyName[128];               // storage area for key name
char lpIniKeyData[1024];              // storage area for data found in, or written to keys
INT INI_KEY_INTEGER = 0;              // storage area for integer value
DWORD INI_DATA_BUFFER_SIZE = 1024;    // size of destination buffer above (lpIniKeyData)
char lpIniDefaultString[] = "\0";     // default string if no data (or key, or section) is found. \0 basically means "empty string"
char *p;                              // used in getting the ini file path

char progname[] = "Dotz screensaver";
HWND hWnd;
HDC pDC;
HBITMAP old;
HBITMAP ourbitmap;
int cx, cy;//screen size
int dot_count, dot_history, inertia, brownian, speed_limit, bounce, chow, starting_pos;
int * framebuf;
int rand_seed, dirty_dot_index;
float dot_fade;
#if show_fps
  int elapsed_frames, fps_start_time;//t:fps
  float frame_rate;//t:fps
#endif

typedef enum { smNone, smConfig, smPassword, smPreview, smSaver } TScrMode;

struct dotz
{
  short r, g, b;
  float x, y, xv, yv;
  struct history
  {
    short x, y;
  };
  struct history hist[dot_historyy];
};
struct dotz *dot;

struct dirty_pixels { int x, y; };
struct dirty_pixels *dirty_dot;

struct tagPOINT InitCursorPos;//initial cursor position

LRESULT CALLBACK winproc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam);
BOOL CALLBACK ConfigDialogProc(HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam);

int rnd() // produces a random number between 0 and 32767
{ rand_seed = rand_seed * 1103515245 + 12345;
  return (unsigned int)(rand_seed / 65536) % 32768;
}

BOOL GetIniFilePath (HINSTANCE hInst)
{ GetModuleFileName(hInst, lpIniFileNameAndPath, sizeof(lpIniFileNameAndPath)); //get application path (C:\DIR\DIR\APP.EXE)
  // Then trim off the program name leaving just the path to the directory (C:\DIR\DIR\)
  p = lpIniFileNameAndPath;
  while (strchr(p, '\\'))  // while there are more separators
  { p = strchr(p, '\\');   // point to next separator
    p++;                   // point to char after last separator
  }
  *p = '\0'; // terminate string
  // append ini file name to the path (C:\DIR\DIR\initest.ini)
  strcat(lpIniFileNameAndPath, lpIniFileName);
  return TRUE;
}

TScrMode GetScrMode(void) //Detirmine the screensaver mode (saver/config/preview/password)
{ TScrMode ScrMode = smNone;
  char *c = GetCommandLine();
  if (*c=='\"') { c++; while (*c != 0 && *c != '\"') c++; }
  else { while (*c != 0 && *c != ' ') c++; }
  if (*c != 0) c++;
  while (*c == ' ') c++;
  if (*c == 0) { ScrMode = smConfig; hWnd=NULL; }
  else
  { if (*c == '-' || *c == '/') c++;
    if (*c == 'p' || *c == 'P' || *c == 'l' || *c == 'L')
    { c++;
      while (*c == ' ' || *c == ':') c++;
      hWnd = (HWND)atoi(c);
      ScrMode = smPreview;
    }
    else if (*c == 's' || *c == 'S') { ScrMode = smSaver; }
    else if (*c == 'c' || *c == 'C')
	{ c++;
      while (*c == ' ' || *c == ':') c++;
	  if (*c == 0) hWnd = GetForegroundWindow();
	  else hWnd = (HWND)atoi(c);
	  ScrMode = smConfig;
    }
    else if (*c == 'a' || *c == 'A')
	{ c++;
      while (*c == ' ' || *c == ':') c++;
      hWnd = (HWND)atoi(c);
	  ScrMode = smPassword;
	}
  }
  return ScrMode;
}

int WINAPI WinMain(HINSTANCE hInst, HINSTANCE hPrevInst, LPSTR lpszCmdLine, int nCmdShow)
{
  WNDCLASSEX winclass;
  MSG msg;
  int i, ii, x, y;
  UINT oldval;
  if (GetScrMode() == smSaver)
  { GetIniFilePath(hInst);
    rand_seed = GetTickCount(); //randomize random seed
    dot_count = GetPrivateProfileInt((LPSTR)strcpy(lpIniSection, "Dotz"), (LPSTR)strcpy(lpIniKeyName, "dot_count"), 0, lpIniFileNameAndPath );
	dot_history = GetPrivateProfileInt((LPSTR)strcpy(lpIniSection, "Dotz"), (LPSTR)strcpy(lpIniKeyName, "dot_history"), 0, lpIniFileNameAndPath );
	if (dot_history > dot_historyy) dot_history = dot_historyy; //limit history to maximum possible
	if (dot_history < 1) dot_history = 1; //limit history to minimum possible
	inertia = GetPrivateProfileInt((LPSTR)strcpy(lpIniSection, "Dotz"), (LPSTR)strcpy(lpIniKeyName, "inertia"), 0, lpIniFileNameAndPath );
	speed_limit = GetPrivateProfileInt((LPSTR)strcpy(lpIniSection, "Dotz"), (LPSTR)strcpy(lpIniKeyName, "speed_limit"), 0, lpIniFileNameAndPath );
	brownian = GetPrivateProfileInt((LPSTR)strcpy(lpIniSection, "Dotz"), (LPSTR)strcpy(lpIniKeyName, "brownian"), 0, lpIniFileNameAndPath );
	bounce = GetPrivateProfileInt((LPSTR)strcpy(lpIniSection, "Dotz"), (LPSTR)strcpy(lpIniKeyName, "bounce"), 0, lpIniFileNameAndPath );
	chow = GetPrivateProfileInt((LPSTR)strcpy(lpIniSection, "Dotz"), (LPSTR)strcpy(lpIniKeyName, "chow"), 0, lpIniFileNameAndPath );
	starting_pos = GetPrivateProfileInt((LPSTR)strcpy(lpIniSection, "Dotz"), (LPSTR)strcpy(lpIniKeyName, "starting_pos"), 0, lpIniFileNameAndPath );
	dot = (struct dotz *)malloc(sizeof(struct dotz)*dot_count); //allocate memory for dotz
	dirty_dot = (struct dirty_pixels *)malloc(sizeof(struct dirty_pixels) * (dot_count * (dot_history + 2))); //allocate memory for dirty pixels
	if ((dot == NULL)||(dirty_dot == NULL))
    { MessageBox(hWnd, "Sorry, there isn't enough memory to run the screensaver.\r\nTry specifying fewer Dotz or shorter trails in the configuration.", "Out of memory", MB_OK );
      free(dot); //free memory allocated for dotz
	  free(dirty_dot); //free memory allocated for dirty pixels
	  return 1;
    }
	cx = GetSystemMetrics(SM_CXSCREEN);
    cy = GetSystemMetrics(SM_CYSCREEN);
    dot_fade = 256.0/dot_history;
    #if show_fps
      fps_start_time = GetTickCount();//t:fps
      elapsed_frames = 0;             //t:fps
      frame_rate = 0;                 //t:fps
	#endif
    for(i = 0; i < dot_count; i++)
    {
      dot[i].r = (rnd() % 256); dot[i].g = (rnd() % 256); dot[i].b = (rnd() % 256);
      if(dot[i].r > 255)dot[i].r = 255;
      if(dot[i].g > 255)dot[i].g = 255;
      if(dot[i].b > 255)dot[i].b = 255;
      if (starting_pos == 0)
      { dot[i].x = cx/2; dot[i].y = cy/2;
        for(ii = 0; ii < dot_history; ii ++) dot[i].hist[ii].x = cx/2;
        for(ii = 0; ii < dot_history; ii ++) dot[i].hist[ii].y = cy/2;
	  }
	  else
	  { x = rnd() % cx; y = rnd() % cy;
		dot[i].x = x; dot[i].y = y;
        for(ii = 0; ii < dot_history; ii ++) dot[i].hist[ii].x = x;
        for(ii = 0; ii < dot_history; ii ++) dot[i].hist[ii].y = y;
      }
      dot[i].xv = 0; dot[i].yv = 0;
    }
    winclass.cbSize = sizeof(WNDCLASSEX);
    winclass.style = CS_HREDRAW | CS_VREDRAW;
    winclass.lpfnWndProc = &winproc;
    winclass.cbClsExtra = 0;
    winclass.cbWndExtra = 0;
    winclass.hInstance = hInst;
    winclass.hIcon = NULL;
    winclass.hCursor = NULL;
    if (chow) winclass.hbrBackground = NULL;
	else winclass.hbrBackground = (HBRUSH)GetStockObject(BLACK_BRUSH);
    winclass.lpszMenuName = NULL;
    winclass.lpszClassName = progname;
    winclass.hIconSm = NULL;
    RegisterClassEx(&winclass);
    hWnd=CreateWindowEx( WS_EX_TOPMOST, progname, progname, WS_POPUP|WS_VISIBLE, 0, 0,
	                     cx, cy, NULL, NULL, hInst, NULL);
    SystemParametersInfo(SPI_SETSCREENSAVEACTIVE, 1, &oldval, 0);
    while (GetMessage(&msg, NULL, 0, 0)) { TranslateMessage(&msg); DispatchMessage(&msg); }
    SystemParametersInfo(SPI_SETSCREENSAVEACTIVE, 0, &oldval, 0);
    free(dot); //free memory allocated for dotz
	free(dirty_dot); //free memory allocated for dirty pixels
	return (msg.wParam);
  }
  else if (GetScrMode() == smConfig) { GetIniFilePath(hInst); DialogBox(hInst, MAKEINTRESOURCE(ID_ConfigDlg), hWnd, ConfigDialogProc); return (msg.wParam); }
  else return 1;
}

void write_int_ini (HWND hWnd, int id, LPSTR key_name)
{ int success, value;
  value = GetDlgItemInt(hWnd, id, &success, 0);
  strcpy (lpIniKeyData, "\0");
  itoa (value, lpIniKeyData, 10);
  WritePrivateProfileString((LPSTR)strcpy(lpIniSection, "Dotz"), key_name,
                            lpIniKeyData, lpIniFileNameAndPath );
}

void write_bool_ini (HWND hWnd, int id, LPSTR key_name)
{ int value;
  if (IsDlgButtonChecked(hWnd, id) == BST_CHECKED) value = 1;
  else value = 0;
  strcpy (lpIniKeyData, "\0");
  itoa (value, lpIniKeyData, 10);
  WritePrivateProfileString((LPSTR)strcpy(lpIniSection, "Dotz"), key_name,
                            lpIniKeyData, lpIniFileNameAndPath );
}

BOOL CALLBACK ConfigDialogProc(HWND hWnd,UINT msg,WPARAM wParam,LPARAM lParam)
{ switch (msg)
  { case WM_INITDIALOG:
    { dot_count = GetPrivateProfileInt((LPSTR)strcpy(lpIniSection, "Dotz"), (LPSTR)strcpy(lpIniKeyName, "dot_count"), 0, lpIniFileNameAndPath );
      SetDlgItemInt(hWnd, ID_Dot_count, dot_count, FALSE);
      dot_history = GetPrivateProfileInt((LPSTR)strcpy(lpIniSection, "Dotz"), (LPSTR)strcpy(lpIniKeyName, "dot_history"), 0, lpIniFileNameAndPath );
	  SetDlgItemInt(hWnd, ID_Dot_history, dot_history, FALSE);
	  inertia = GetPrivateProfileInt((LPSTR)strcpy(lpIniSection, "Dotz"), (LPSTR)strcpy(lpIniKeyName, "inertia"), 0, lpIniFileNameAndPath );
	  SetDlgItemInt(hWnd, ID_Inertia, inertia, FALSE);
  	  speed_limit = GetPrivateProfileInt((LPSTR)strcpy(lpIniSection, "Dotz"), (LPSTR)strcpy(lpIniKeyName, "speed_limit"), 0, lpIniFileNameAndPath );
	  SetDlgItemInt(hWnd, ID_Speed_limit, speed_limit, FALSE);
   	  brownian = GetPrivateProfileInt((LPSTR)strcpy(lpIniSection, "Dotz"), (LPSTR)strcpy(lpIniKeyName, "brownian"), 0, lpIniFileNameAndPath );
      CheckDlgButton(hWnd, 108, brownian);
	  bounce = GetPrivateProfileInt((LPSTR)strcpy(lpIniSection, "Dotz"), (LPSTR)strcpy(lpIniKeyName, "bounce"), 0, lpIniFileNameAndPath );
      CheckDlgButton(hWnd, 109, bounce);
   	  chow = GetPrivateProfileInt((LPSTR)strcpy(lpIniSection, "Dotz"), (LPSTR)strcpy(lpIniKeyName, "chow"), 0, lpIniFileNameAndPath );
      CheckDlgButton(hWnd, 110, chow);
   	  starting_pos = GetPrivateProfileInt((LPSTR)strcpy(lpIniSection, "Dotz"), (LPSTR)strcpy(lpIniKeyName, "starting_pos"), 0, lpIniFileNameAndPath );
      CheckDlgButton(hWnd, 111, starting_pos);
      break;
    }
    case WM_COMMAND:
      switch (LOWORD(wParam))
	  { case ID_OK:
          write_int_ini(hWnd, 104, (LPSTR)strcpy(lpIniKeyName, "dot_count"));
          write_int_ini(hWnd, 105, (LPSTR)strcpy(lpIniKeyName, "dot_history"));
		  write_int_ini(hWnd, 106, (LPSTR)strcpy(lpIniKeyName, "inertia"));
		  write_int_ini(hWnd, 107, (LPSTR)strcpy(lpIniKeyName, "speed_limit"));
		  write_bool_ini(hWnd, 108, (LPSTR)strcpy(lpIniKeyName, "brownian"));
		  write_bool_ini(hWnd, 109, (LPSTR)strcpy(lpIniKeyName, "bounce"));
		  write_bool_ini(hWnd, 110, (LPSTR)strcpy(lpIniKeyName, "chow"));
		  write_bool_ini(hWnd, 111, (LPSTR)strcpy(lpIniKeyName, "starting_pos"));
		  EndDialog(hWnd, LOWORD(wParam));
		  break;
        case ID_About:
		  MessageBox(hWnd, "Dotz screensaver\r\nv0.8\r\n31/07/2000\r\n\r\nby Stefan Ferreira\r\nStefan_F@SoftHome.net\r\nhttp://www.geocities.com/dragonhawk777/\r\n\r\nThis screensaver is free, but I would\r\ngreatly appreciate any donations.", "About", MB_OK );
		  break;
		case ID_Cancel:
          EndDialog(hWnd, LOWORD(wParam));
		  break;
      }
	  break;
  }
  return FALSE;
}



void put_pixel(HDC hDC, int * framebuf, int x_loc, int y_loc, int r_val, int g_val, int b_val)
{
  *(framebuf+(y_loc*cx + x_loc)) = RGB(r_val, g_val, b_val);
//  BitBlt(hDC,x_loc,y_loc,1,1,pDC,x_loc,y_loc,SRCCOPY); //use if dirty rectangles isn't used
  dirty_dot[dirty_dot_index].x = x_loc;
  dirty_dot[dirty_dot_index].y = y_loc;
  dirty_dot_index++; //dirty rectangles method
}

void render(HDC hDC, int * framebuf)
{
  int i, ii, r, g, b;
  dirty_dot_index = 0;
  for (i = 0; i < dot_count; i++) {
    for (ii = (dot_history-1); ii > 0; ii--) {
      dot[i].hist[ii].x = dot[i].hist[ii-1].x;
      dot[i].hist[ii].y = dot[i].hist[ii-1].y;
      r = dot[i].r - ((ii+1) * dot_fade); if(r < 0)r = 0; //4-16-256/history
      g = dot[i].g - ((ii+1) * dot_fade); if(g < 0)g = 0;
      b = dot[i].b - ((ii+1) * dot_fade); if(b < 0)b = 0;
      put_pixel(hDC, framebuf, dot[i].hist[ii].x, dot[i].hist[ii].y, r, g, b);
    }
    put_pixel(hDC, framebuf, dot[i].hist[(dot_history-1)].x, dot[i].hist[(dot_history-1)].y, 0, 0, 0);
    dot[i].hist[0].x = dot[i].x;
    dot[i].hist[0].y = dot[i].y;
    r = dot[i].r - dot_fade; if(r < 0)r = 0;
    g = dot[i].g - dot_fade; if(g < 0)g = 0;
    b = dot[i].b - dot_fade; if(b < 0)b = 0;
    put_pixel(hDC, framebuf, dot[i].hist[0].x, dot[i].hist[0].y, r, g, b);
    if (brownian)
    { dot[i].xv = (rnd() % 3) - 1;
      dot[i].yv = (rnd() % 3) - 1;
    }
    else//(32767*2.5)
	{ dot[i].xv = dot[i].xv + ((float)((rnd()-16383)%1001) / inertia);
      dot[i].yv = dot[i].yv + ((float)((rnd()-16383)%1001) / inertia);
      if(dot[i].xv > speed_limit) dot[i].xv = speed_limit;
      if(dot[i].xv < -speed_limit) dot[i].xv = -speed_limit;
      if(dot[i].yv > speed_limit) dot[i].yv = speed_limit;
      if(dot[i].yv < -speed_limit) dot[i].yv = -speed_limit;
    }
    dot[i].x = dot[i].x + dot[i].xv;
    dot[i].y = dot[i].y + dot[i].yv;
    if (bounce)
	{ if(dot[i].x < 0){ dot[i].x = 0; dot[i].xv = 0 - dot[i].xv; }
      if(dot[i].x > (cx-1)){ dot[i].x = (cx-1); dot[i].xv = 0 - dot[i].xv; }
      if(dot[i].y < 0){ dot[i].y = 0; dot[i].yv = 0 - dot[i].yv; }
      if(dot[i].y > (cy-1)){ dot[i].y = (cy-1); dot[i].yv = 0 - dot[i].yv; }
    }
	else
	{ if(dot[i].x < 0) dot[i].x = (cx-1);
      if(dot[i].x > (cx-1)) dot[i].x = 0;
      if(dot[i].y < 0) dot[i].y = (cy-1);
      if(dot[i].y > (cy-1)) dot[i].y = 0;
    }
    put_pixel(hDC, framebuf, dot[i].x, dot[i].y, dot[i].r, dot[i].g, dot[i].b);
  }
  for (i = 0; i < dot_count*(dot_history+2); i++)//dirty rectangles method
    BitBlt(hDC,dirty_dot[i].x,dirty_dot[i].y,1,1,pDC,dirty_dot[i].x,dirty_dot[i].y,SRCCOPY);
}

void init_framebuf(void)
{ HDC hDC;
  BITMAPINFO bitmapinfo;
  hDC = CreateCompatibleDC(NULL);
  bitmapinfo.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
  bitmapinfo.bmiHeader.biWidth = cx;
  bitmapinfo.bmiHeader.biHeight = 0 - cy; // top-down
  bitmapinfo.bmiHeader.biPlanes = 1;
  bitmapinfo.bmiHeader.biBitCount = 32;
  bitmapinfo.bmiHeader.biCompression = BI_RGB;
  bitmapinfo.bmiHeader.biSizeImage = 0;
  bitmapinfo.bmiHeader.biClrUsed = 0;
  bitmapinfo.bmiHeader.biClrImportant = 0;
  ourbitmap = CreateDIBSection(hDC, &bitmapinfo, DIB_RGB_COLORS, &framebuf, 0, 0);
  pDC = CreateCompatibleDC(NULL);
  old = SelectObject(pDC, ourbitmap);
  DeleteDC(hDC);
}

void deinit_framebuf(void) { SelectObject(pDC, old); DeleteDC(pDC); DeleteObject(ourbitmap); }

LRESULT CALLBACK winproc(HWND hWnd, UINT uMsg, WPARAM wParam, LPARAM lParam)
{
  HDC hDC;
  PAINTSTRUCT PtStr;
  #if show_fps
    int fps_elapsed_time;//t:fps
    char stringA[11];    //t:fps
  #endif
  POINT CursorPos;
  switch (uMsg) {
  case WM_DESTROY:
	deinit_framebuf();
    PostQuitMessage(0);
    KillTimer(hWnd, 1);
	return 0;
    break;
  case WM_CLOSE:
    deinit_framebuf();
    PostQuitMessage(0);
    KillTimer(hWnd, 1);
	return 0;
    break;
  case WM_CREATE:
    SetTimer(hWnd, 1, 10, NULL); //10 miliseconds limits framerate < 100fps
	GetCursorPos(&InitCursorPos);
	SetCursor(NULL);
    init_framebuf();
    break;
  case WM_TIMER:
    InvalidateRgn(hWnd, 0, 0);
    UpdateWindow (hWnd);
    break;
  case WM_PAINT:
    hDC=BeginPaint(hWnd,&PtStr);
    render(hDC, framebuf);
	#if show_fps
	  elapsed_frames++;                                            //t:fps
	  fps_elapsed_time = GetTickCount() - fps_start_time;          //t:fps
	  if (fps_elapsed_time >= 3000)                                //t:fps
	  { frame_rate = ((float)elapsed_frames/fps_elapsed_time)*1000;//t:fps
	    elapsed_frames = 0;                                        //t:fps
        fps_start_time = GetTickCount();                           //t:fps
      }                                                            //t:fps
	  sprintf(stringA, "%f", (float)frame_rate);                   //t:fps
	  stringA[8] = 'f'; stringA[9] = 'p'; stringA[10] = 's';       //t:fps
	  TextOut(hDC, 0, 0, (LPCTSTR) stringA, 11);                   //t:fps
    #endif
	EndPaint(hWnd, &PtStr);
    break;
  case WM_LBUTTONDOWN: case WM_MBUTTONDOWN: case WM_RBUTTONDOWN: case WM_KEYDOWN:
	PostMessage(hWnd, WM_CLOSE, 0, 0);
    break;
  case WM_ACTIVATE: case WM_ACTIVATEAPP: case WM_NCACTIVATE:
	if(LOWORD(wParam) == WA_INACTIVE) PostMessage(hWnd, WM_CLOSE, 0, 0);
	break;
  case WM_MOUSEMOVE:
    GetCursorPos(&CursorPos);
	if ((abs(CursorPos.x - InitCursorPos.x) > 5)||(abs(CursorPos.y - InitCursorPos.y) > 5))
      PostMessage(hWnd, WM_CLOSE, 0, 0);
	break;
  }
  return DefWindowProc (hWnd, uMsg, wParam, lParam);//return 0;
}
